import os
import tensorflow as tf
os.environ['TF_GPU_ALLOCATOR'] = 'cuda_malloc_async'
os.environ['TF_FORCE_GPU_ALLOW_GROWTH'] = 'true'
gpus = tf.config.experimental.list_physical_devices("GPU")


# TODO: Create a method for this for all files to share
for gpu in gpus:
    # Restrict TensorFlow to only allocate 8GB of memory on the first GPU
    try:
        tf.config.experimental.set_virtual_device_configuration(
            gpu,
            [tf.config.experimental.VirtualDeviceConfiguration(memory_limit=8192)])
    except RuntimeError as e:
        # Virtual devices must be set before GPUs have been initialized
        print(e)

logical_gpus = tf.config.experimental.list_logical_devices('GPU')
print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")

import numpy as np
import autograd.numpy as anp
__requires__ = 'pymoo==0.5.0'
import pkg_resources
pkg_resources.require("pymoo==0.5.0")
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.optimize import minimize
from src.pixel_change_minimise import PixelChangeMinimise
import src.de_mutations as de_mutation
import src.callbacks as callback
import src.eliminations as eliminate
import src.crossovers as crossover
import src.sampling as sampling
import random
import time
from src import out

# MyProblem - Extends ElementwiseProblem, contains methods for our problem.
# image - Path to the image we're evaluating.
# model - Defines the ML prediction model. Defaults to vgg16.
class DecreaseBestModelPrediction(PixelChangeMinimise):
    def __init__(self,
                 image="",
                 model=None):
        # model = vgg19.VGG19(weights='imagenet')):
        super().__init__(image, model)

    """
    _evaluate - Simple at the moment, gets the latest predictions and stores them.
    return - Returns "F" - top value and "G" Array of all predicted values.
    """
    def _evaluate(self, child, out, *args, **kwargs):
        values = self.get_predictions(child[1])
        values.append(child[1])
        values.append(child[2])
        f1 = 0

        # if the top-1 prediction is identical to the original image we want to:
        # 1. Reduce the confidence for the top-1 prediction
        # 2. Increase the confidence for the top-2 prediction
        if values[0][0] == self.original_classification[0][0]:
            # out["F"] = np.array(values[0][2] - values[1][2], dtype=float)
            f1 = np.array(values[0][2] - values[1][2], dtype=float)
            values.append(False)
        else:
            # if the top-1 prediction is not identical to the original image we want to
            # simply increase its corresponding confidence
            # out["F"] = -np.array(values[0][2], dtype=float)
            f1 = -np.array(values[0][2], dtype=float)
            values.append(True)

        total_noise = 0
        for mutation in child[2]:
            total_noise += mutation[3]

        out["F"] = anp.column_stack([f1, len(child[2]), total_noise])
        out["G"] = values


def run_mutations(
        pop_size=50,
        n_gen=50,
        image="",
        model=None,
        directory="output_data/",
        seed=random.seed(time.time())):

    # Tensorflow information for debugging
    print("Version: ", tf.__version__)
    print("Eager mode: ", tf.executing_eagerly())
    print("GPU is", "available" if tf.config.experimental.list_physical_devices("GPU") else "NOT AVAILABLE")

    # Create our algorithm based off NSGA2.
    algorithm = NSGA2(pop_size=pop_size,
                      sampling=sampling.InitialImageSampling(),
                      crossover=crossover.ArraySlice(),
                      mutation=de_mutation.PixelMutation(),
                      eliminate_duplicates=eliminate.SimpleElimination())

    problem = DecreaseBestModelPrediction(image, model)
    # Use minimize, giving our problem and the GA algorithm.
    res = minimize(problem,
                   algorithm,
                   ('n_gen', n_gen),
                   seed=seed,
                   callback=callback.PredictionsAndImage(),
                   verbose=True)

    out.output_data(problem, res, directory)
