import os
import random
import time

import numpy as np
import tensorflow as tf
from pymoo.algorithms.soo.nonconvex.de import DE
from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.operators.sampling.lhs import LHS
from pymoo.optimize import minimize

import src.callbacks as callback
import src.crossovers as crossover
import src.eliminations as eliminate
import src.de_mutations as de_mutation
import src.sampling as sampling
from src import out
from src.pixel_change_minimise import PixelChangeMinimise

os.environ['TF_GPU_ALLOCATOR'] = 'cuda_malloc_async'
os.environ['TF_FORCE_GPU_ALLOW_GROWTH'] = 'true'
gpus = tf.config.list_physical_devices('GPU')
for gpu in gpus:
    # Restrict TensorFlow to only allocate 8GB of memory on the first GPU
    try:
        tf.config.experimental.set_virtual_device_configuration(
            gpu,
            [tf.config.experimental.VirtualDeviceConfiguration(memory_limit=8192)])
    except RuntimeError as e:
        # Virtual devices must be set before GPUs have been initialized
        print(e)

logical_gpus = tf.config.experimental.list_logical_devices('GPU')
print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")


# MyProblem - Extends ElementwiseProblem, contains methods for our problem.
# image - Path to the image we're evaluating.
# model - Defines the ML prediction model. Defaults to vgg16.
class DecreaseBestModelPrediction(PixelChangeMinimise):
    def __init__(self,
                 image="",
                 model=None):
        super().__init__(image, model)

    """
    _evaluate - Simple at the moment, gets the latest predictions and stores them.
    return - Returns "F" - top value and "G" Array of all predicted values.
    """
    def _evaluate(self, child, out, *args, **kwargs):
        values = self.get_predictions(child[1])
        values.append(child[1])
        values.append(child[2])

        # if the top-1 prediction is identical to the original image we want to:
        # 1. Reduce the confidence for the top-1 prediction
        # 2. Increase the confidence for the top-2 prediction
        if values[0][0] == self.original_classification[0][0]:
            out["F"] = np.array(values[0][2] - values[1][2], dtype=float)
            values.append(False)
        else:
            # if the top-1 prediction is not identical to the original image we want to
            # simply increase its corresponding confidence
            out["F"] = -np.array(values[0][2], dtype=float)
            values.append(True)

        out["G"] = values


def run_mutations(
        pop_size=50,
        n_gen=50,
        image="",
        model=None,
        directory="output_data_soo/",
        seed=random.seed(time.time())):

    # Tensorflow information for debugging
    print("Version: ", tf.__version__)
    print("Eager mode: ", tf.executing_eagerly())
    print("GPU is", "available" if tf.config.experimental.list_physical_devices("GPU") else "NOT AVAILABLE")

    # Create our algorithm based off of GA.
    algorithm = GA(pop_size=pop_size,
                   sampling=sampling.InitialImageSampling(),
                   crossover=crossover.ArraySlice(),
                   mutation=de_mutation.PixelMutation(),
                   eliminate_duplicates=eliminate.SimpleElimination())

    problem = DecreaseBestModelPrediction(image, model)
    # Use minimize, giving our problem and the GA algorithm. 100 generations.
    res = minimize(problem,
                   algorithm,
                   ('n_gen', n_gen),
                   seed=seed,
                   callback=callback.PredictionsAndImage(),
                   verbose=True)

    out.output_data(problem, res, directory)
